{% set title = title|default(_fsdomain("Reauthenticate")) %}
{% extends "security/base.html" %}
{% from "security/_macros.html" import render_field_with_errors, render_field, prop_next, render_csrf %}

{% block content %}
  {% include "security/_messages.html" %}
  <h1>{{ _fsdomain("Reauthenticate") }}</h1>
  <form action="{{ url_for_security('verify') }}{{ prop_next() }}" method="post" name="verify_form">
    {{ verify_form.hidden_tag() }}
    {{ render_field_with_errors(verify_form.password) }}
    {{ render_field(verify_form.submit) }}
  </form>
  {% if has_webauthn_verify_credential %}
    <hr class="fs-gap">
    <h2>{{ _fsdomain("Use a Passkey to Reauthenticate") }}</h2>
    <form action="{{ url_for_security('wan_verify') }}{{ prop_next() }}" method="post" name="wan_verify_form">
      {# explicitly render csrf_token so we can change the ID so we don't get duplicates #}
      {{ render_csrf(wan_verify_form, "wan") }}
      {{ render_field(wan_verify_form.submit) }}
    </form>
  {% endif %}
{% endblock content %}
