{% set title = title|default(_fsdomain("Select Two-Factor Method")) %}
{% extends "security/base.html" %}
{% from "security/_macros.html" import prop_next, render_field_with_errors, render_field %}

{% block content %}
  {% include "security/_messages.html" %}
  <h1>{{ _fsdomain('Select Two-Factor Method') }}</h1>
  <form action="{{ url_for_security('tf_select') }}{{ prop_next() }}" method="post" name="tf_select">
    {{ two_factor_select_form.hidden_tag() }}
    {{ render_field_with_errors(two_factor_select_form.which) }}
    {{ render_field(two_factor_select_form.submit) }}
  </form>
  {% include "security/_menu.html" %}
{% endblock content %}
